package be.iminds.ilabt.jfed.rspec.rspec_source;

import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.imutable_impl.ImmutableModelRspec;
import be.iminds.ilabt.jfed.rspec.model.imutable_impl.ImmutableRspecFactory;
import be.iminds.ilabt.jfed.rspec.parser.RspecCompare;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.StaxRspecParser;
import be.iminds.ilabt.jfed.rspec.util.ProgressHandler;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.ThreadFactoryUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.stream.XMLStreamException;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

/**
 * ImmutableRspecSource
 * <p>
 * Note: this is a fully immutable class, except for some caching
 *
 * This is implemented by never storing any BasicRSpecModel or FXRspecModel.
 * We only store ImmutableModelRspec
 */
public abstract class ImmutableRspecSource implements RspecSource {
    private static final Logger LOG = LoggerFactory.getLogger(ImmutableRspecSource.class);

    protected final Map<ModelRspec.RequestRspecSpecialCases, StringRspec> stringRspecBySpecialCase =
            new EnumMap<>(ModelRspec.RequestRspecSpecialCases.class);
    @Nullable protected ImmutableModelRspec immutableModelRspec = null;

    protected final boolean isXmlBased;

    @Nonnull private final ModelRspecType defaultMutableModelRspecType;

    /* COPY constructor */
    protected ImmutableRspecSource(@Nonnull RspecSource orig) {
        if (orig.isXmlBased()) {
            this.stringRspecBySpecialCase.put(ModelRspec.RequestRspecSpecialCases.NONE, orig.getStringRspec());
            this.isXmlBased = true;
            this.defaultMutableModelRspecType = ModelRspecType.BASIC;
        } else {
            ImmutableModelRspec origModelRspec = orig.getImmutableModelRspec();
            assert origModelRspec != null;
            this.immutableModelRspec = ImmutableRspecFactory.getInstance().copyModelRspec(origModelRspec, null); //not really needed to copy, but we'll do it anyway
            this.defaultMutableModelRspecType = ModelRspecType.BASIC;
            this.isXmlBased = false;
        }
    }

    protected ImmutableRspecSource(@Nonnull StringRspec stringRspec) {
        this(stringRspec, ModelRspecType.BASIC);
    }

    /**
     * @param defaultMutableModelRspecType default ModelRspecType to use. may be null. (= BASIC in that case)
     */
    protected ImmutableRspecSource(@Nonnull StringRspec stringRspec, @Nonnull ModelRspecType defaultMutableModelRspecType) {
        assert stringRspec != null;
        this.stringRspecBySpecialCase.put(ModelRspec.RequestRspecSpecialCases.NONE, stringRspec);
        this.isXmlBased = true;
        this.defaultMutableModelRspecType = defaultMutableModelRspecType;
    }

    protected ImmutableRspecSource(@Nonnull ModelRspec modelRspec) {
        assert modelRspec != null;
        this.immutableModelRspec = ImmutableRspecFactory.getInstance().copyModelRspec(modelRspec, null);
        this.defaultMutableModelRspecType = modelRspec.getModelRspecType();
        this.isXmlBased = false;
    }

    /**
     * @param defaultMutableModelRspecType default ModelRspecType to use. may be null. (= BASIC in that case)
     */
    public ImmutableRspecSource(@Nonnull String xmlRspecString, @Nonnull ModelRspecType defaultMutableModelRspecType) {
        this(new StringRspec(xmlRspecString), defaultMutableModelRspecType);
    }

    @Override
    public boolean isXmlBased() {
        return isXmlBased;
    }

    @Override
    @Nullable
    public String getRspecXmlString() {
        return getRspecXmlString(ModelRspec.RequestRspecSpecialCases.NONE);
    }

    @Override
    @Nullable
    public String getRspecXmlString(@Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase) {
        final StringRspec stringRspec = getStringRspec(rspecSpecialCase);
        if (stringRspec != null)
            return stringRspec.getXmlString();
        else return null;
    }

    protected boolean parseFailed = false;

    @Nonnull
    protected String modelToString(@Nonnull ModelRspec modelRspec, @Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase) {
        assert modelRspec != null;
        return modelRspec.toGeni3Rspec(rspecSpecialCase);
    }


    /**
     * @return a modelRspec, created by any mutable RspecFactory
     */
    @Override
    @Nullable
    public ModelRspec getMutableModelRspec() {
        return getModelRspec(defaultMutableModelRspecType, (ProgressHandler[]) null);
    }

    /**
     * @return a modelRspec, created by any immutable RspecFactory
     */
    @Override
    @Nullable
    public ImmutableModelRspec getImmutableModelRspec() {
        if (immutableModelRspec == null) {
            immutableModelRspec = getImmutableModelRspec((ProgressHandler[]) null);
        }
        return immutableModelRspec;
    }

    /**
     * @return a modelRspec, created by any mutable RspecFactory
     */
    @Override
    @Nullable
    public ModelRspec getMutableModelRspec(ProgressHandler... progressHandlers) {
        return getModelRspec(defaultMutableModelRspecType, progressHandlers);
    }

    /**
     * @return a modelRspec, created by any Immutable RspecFactory
     */
    @Override
    @Nullable
    public ImmutableModelRspec getImmutableModelRspec(ProgressHandler... progressHandlers) {
        if (immutableModelRspec == null) {
            immutableModelRspec = getModelRspec(ModelRspecType.IMMUTABLE, (ProgressHandler[]) null);
        }
        return immutableModelRspec;
    }

    @Override
    public boolean isModelRspecAvailable(@Nonnull ModelRspecType modelRspecType) {
        return modelRspecType == ModelRspecType.IMMUTABLE && immutableModelRspec != null;
    }

    @SuppressWarnings("unchecked")
    @Override
    @Nullable
    public <M extends ModelRspec> M getModelRspec(@Nonnull ModelRspecType modelRspecType, ProgressHandler... progressHandlers) {
        if (immutableModelRspec != null && modelRspecType == ModelRspecType.IMMUTABLE) {
            return (M)immutableModelRspec;
        }
        if (parseFailed)
            return null;

        if (immutableModelRspec != null) {
            //directly convert immutableModelRspec to the wanted type
            return (M) RspecFactoryFactory.getRspecFactoryInstance(modelRspecType).copyModelRspec(immutableModelRspec, null);
        }


        try {
            ModelRspec res = parseStringToModelRspec(modelRspecType, progressHandlers);

            if (res != null && immutableModelRspec == null) {
                assert res.getModelRspecType() == modelRspecType;
                //cache immutableModelRspec
                if (modelRspecType == ModelRspecType.IMMUTABLE) {
                    immutableModelRspec = (ImmutableModelRspec) res;
                } else {
                    immutableModelRspec = ImmutableRspecFactory.getInstance().copyModelRspec(res, null);
                }
                assert immutableModelRspec.getModelRspecType() == ModelRspecType.IMMUTABLE;
            }

            return (M)res;
        } catch (RspecParseException e) {
            lastInvalidRspecException = (RspecParseException) e;
            LOG.warn("RSpec String parse failed. Will not return ModelRspec", e);
            parseFailed = true;
            return null;
        }
    }

    @Nullable
    private ModelRspec parseStringToModelRspec(@Nonnull ModelRspecType modelRspecType, ProgressHandler... progressHandlers) throws RspecParseException {
        //get the correct StringRspec
        StringRspec stringRspec;
        if (stringRspecBySpecialCase.containsKey(ModelRspec.RequestRspecSpecialCases.NONE))
            stringRspec = stringRspecBySpecialCase.get(ModelRspec.RequestRspecSpecialCases.NONE);
        else {
            if (!stringRspecBySpecialCase.isEmpty())
                stringRspec = stringRspecBySpecialCase.get(stringRspecBySpecialCase.keySet().iterator().next());
            else {
                //this only occurs when we do not have a StringRspec, but we have a ModelRspec of the wrong type
                stringRspec =
                        getStringRspec(); //no stringRspec stored. getStringRspec() will generate one from the modelRspec
                //this can be done more efficiently, as ModelRspec can be copied instead of re-parsed.
                //   But this is a safe method, and we can replace later if needed.
            }
        }

        if (stringRspec == null)
            return null;

        String xmlString = stringRspec.getXmlString();

        //quick HACK to allow the planetlab Europe rspec manifest
        if (xmlString.contains("<RSpec")) {
            LOG.debug("HACK: rewriting received planetlab manifest XML to make it parsable xml");
            xmlString =
                    xmlString.replace("<RSpec type=\"SFA\"", "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" type=\"manifest\"");
            xmlString =
                    xmlString.replace("<RSpec ", "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" type=\"manifest\"");
            xmlString = xmlString.replace("</RSpec>", "</rspec>");
            xmlString = xmlString.replace("<network name=\"ple\">", "");
            xmlString = xmlString.replace("</network>", "");
            LOG.debug("HACK: rewritten xml: " + xmlString);
        }

        RspecFactory factory = RspecFactoryFactory.getRspecFactoryInstance(modelRspecType);
        if (modelRspecType == ModelRspecType.IMMUTABLE) {
            //cannot parse to immutable
            factory =  RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.BASIC);
        }
        StaxRspecParser parser = new StaxRspecParser(xmlString, factory);

        //add progress-handlers
        if (progressHandlers != null)
            for (ProgressHandler handler : progressHandlers)
                if (handler != null)
                    parser.addProgressHandler(handler);

        ModelRspec res = parser.call();
        if (res != null && modelRspecType == ModelRspecType.IMMUTABLE) {
            //convert mutable to immutable after parse
            res = RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.IMMUTABLE).copyModelRspec(res, null);
        }
        return res;
    }


    private RspecParseException lastInvalidRspecException = null;

    @Override
    @Nullable
    public RspecParseException getLastInvalidRspecException() {
        return lastInvalidRspecException;
    }

    @Override
    @Nullable
    public StringRspec getStringRspec() {
        return getStringRspec(ModelRspec.RequestRspecSpecialCases.NONE);
    }

    @Override
    @Nullable
    public StringRspec getStringRspec(@Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase) {
        if (stringRspecBySpecialCase.containsKey(rspecSpecialCase))
            return stringRspecBySpecialCase.get(rspecSpecialCase);
        ModelRspec modelRspec = getImmutableModelRspec();
        if (modelRspec == null) {
            return null;
        }

        String rspecXml = modelToString(modelRspec, rspecSpecialCase);
        StringRspec stringRspec = new StringRspec(rspecXml);
        stringRspecBySpecialCase.put(rspecSpecialCase, stringRspec);
        return stringRspec;
    }


    @Override
    @Nonnull
    public List<GeniUrn> getAllComponentManagerUrns() {
        ModelRspec model = getImmutableModelRspec();
        if (model != null) {
            List<GeniUrn> res = model.getAllComponentManagerUrns();

            StringRspec stringRspec = getStringRspec();

            if (stringRspec == null)
                return Collections.emptyList();

            assert stringRspec.isWellFormed();

            //parse generated rspec for openflow component URNs.
            //   Note: This could be done more efficiently, by parsing "others", but it's not worth the trouble

            for (String urnString : StringRspec.parseOpenflowComponentManagers(stringRspec.getXmlString())) {
                GeniUrn geniUrn = GeniUrn.parse(urnString);

                if (geniUrn != null)
                    res.add(geniUrn);
                else {
                    LOG.warn("Could not parse openflowComponentManager-urn {}", urnString);
                }
            }

            return res;
        } else {
            StringRspec stringRspec = getStringRspec();
            if (stringRspec != null && stringRspec.isWellFormed()) {
                List<GeniUrn> res = new ArrayList<>();
                for (String urnString : stringRspec.getAllComponentManagerUrns()) {
                    GeniUrn geniUrn = GeniUrn.parse(urnString);

                    if (geniUrn != null)
                        res.add(geniUrn);
                    else {
                        LOG.warn("Could not parse openflowComponentManager-urn {}", urnString);
                    }
                }

                return res;
            } else
                return Collections.emptyList();
        }
    }

    @Override
    public List<BasicStringRspec.BasicNodeInfo> getBasicNodeInfo() {
        ModelRspec model = getImmutableModelRspec();

        if (model != null) {
            List<BasicStringRspec.BasicNodeInfo> res = new ArrayList<>();
            List<? extends RspecNode> modelNodes = model.getNodes();

            for (RspecNode node : modelNodes) {
                List<String> sliverTypeNames = new ArrayList<>();
                for (SliverType sliverType : node.getSliverTypes())
                    sliverTypeNames.add(sliverType.getName());
                BasicStringRspec.BasicNodeInfo ni = new BasicStringRspec.BasicNodeInfo(
                        node.getClientId(),
                        node.getComponentId() == null ? null : node.getComponentId().toString(),
                        node.getComponentManagerId() == null ? null : node.getComponentManagerId().toString(),
                        node.getAvailable(),
                        node.getExclusive(),
                        sliverTypeNames,
                        node.getAnsibleGroups());
                res.add(ni);
            }

            return res;
        } else {
            StringRspec stringRspec = getStringRspec();
            LOG.debug("The RSpec could not be parsed using ModelRspec, will try StringRspec method.");
            if (stringRspec != null && stringRspec.isWellFormed()) {
                return stringRspec.getBasicNodeInfo();
            } else {
                LOG.warn("The RSpec could not be parsed using any method. Will return null.");
                return null;
            }
        }
    }

    private final Map<ModelRspecType, LosingDataInfo> losingDataInfoMap = new HashMap<>();

    @Nullable
    private LosingDataInfo getLosingDataInfo(ModelRspecType modelRspecType) {
        if (!isXmlBased)
            return null; //cannot lose data when not xml-based!

        if (!losingDataInfoMap.containsKey(modelRspecType)) {
            ModelRspec modelRspec = getModelRspec(modelRspecType);
            String stringRspec = getRspecXmlString();

            if (modelRspec == null || stringRspec == null)
                return null;

            LosingDataInfo losingDataInfo;
            //set isLosingData, but handle any failure.
            try {
                RspecCompare compareWithOrig =
                        new RspecCompare(modelRspec.toGeni3Rspec(), stringRspec, "the converted RSpec", "the original RSpec");
                losingDataInfo =
                        new LosingDataInfo(!compareWithOrig.isEqual(), compareWithOrig.getDifferences()); //TODO: only calculate and set losingData this when it is actually needed
            } catch (XMLStreamException ex) {
                LOG.error("Exception while using RspecCompare at end of parse. location=" + RspecParseException.locationToString(ex.getLocation()), ex);
                losingDataInfo = new LosingDataInfo(true, /*"Exception while comparing: "+ex.getMessage()*/null);
            } catch (AssertionError | Exception ex) {
                LOG.error("Exception while using RspecCompare at end of parse. location=unknown", ex);
                losingDataInfo = new LosingDataInfo(true, /*"Exception while comparing: "+ex.getMessage()*/null);
            }

            losingDataInfoMap.put(modelRspecType, losingDataInfo);
        }

        return losingDataInfoMap.get(modelRspecType);
    }

    @Override
    public boolean isLosingData(ModelRspecType modelRspecType) {
        LosingDataInfo losingDataInfo = getLosingDataInfo(modelRspecType);
        return losingDataInfo != null && losingDataInfo.isLosingData();
    }

    @Override
    public List<RspecCompare.Difference> getLosingDataDifferences(ModelRspecType modelRspecType) {
        LosingDataInfo losingDataInfo = getLosingDataInfo(modelRspecType);
        if (losingDataInfo != null)
            return losingDataInfo.getLosingDataDifferences();
        else
            return null;
    }
}
